function [lambda,Phi,order] = omassi(Y,Fs,varargin)
%%SSI Covariance-driven Stochastic Subspace Identification (output only).
% Use like this: [lambda,Phi] = omassi(Y,Fs,...)
%
% *Y is matrix with measurement data for channels (each channel is one
% column)
%
% *Fs is the sampling frequency corresponding to the data matrix
%
% Optional parameters:
%
% Key/name:         Comment:                                Standard value:
% ------------------------------------------------------------------------
% expected_order    Expected order of system (n)            [] (prompted)
% or
% blockrows                                                 [] (calculated
%                                                      from expected_order)
%
% order             System orders (for stab. diagram)       [] (prompted)
% 
%
% Knut Andreas Kvle, 2014
%

p=inputParser;
addParameter(p,'expected_order',[],@isnumeric);      %expected number of modes
addParameter(p,'order',[],@isnumeric);   %system orders (for stabilization diagram etc.)
addParameter(p,'blockrows',[]);

parse(p,varargin{:})
n_exp=p.Results.expected_order;
n=p.Results.order;
i=p.Results.blockrows;

dt=1/Fs;
[N,l] = size(Y);

%% DECIDE ORDERS, BY PLOTTING TRACE OF PSD+
if isempty(n) && isempty(n_exp)
    ZP=1;
    tWelch=60;  %seconds per division
    Nwelch=max(round(t(end)/tWelch),1);
    S=0;
    for channel=1:size(Y,2)
        [S0,f] = welchAPSD(Y(:,channel),dt,Nwelch,ZP,'Hz');
        S=S+S0;
    end
    
    plot(f,S)
    xlabel('f [Hz]')
    ylabel('Trace of PSD^+')
    
    n_exp = input('Please suggest number of structural modes = ');
end

%% SUGGESTED VALUE FOR i (number of Block rows in Toeplitz matrix)
if isempty(i)
    if isempty(n_exp)
        i=2*ceil(max(n)/l);
    else
        i=ceil(8*(n_exp/l));
    end
end

disp(['*** Number of block rows i = ' num2str(i)]);

%% ESTABLISH TOEPLITZ AND LAG-SHIFTED TOEPLITZ MATRIX
T=zeros(i*l,i*l);
T2=zeros(i*l,i*l);
range0=1:l;

for row=1:i            %go through all block rows from 1 to i
    for col=1:i        %go through all block columns from 1 to i
        %Toeplitz
        irs=i+(row-1)-(col-1);
        T(range0+(row-1)*l,range0+(col-1)*l) = 1/(N-irs) * Y(1:N-irs,:).' * Y(irs+1:N,:);
        
        %Lag-shifted Toeplitz
        irs2=i+row-(col-1);
        T2(range0+(row-1)*l,range0+(col-1)*l) = 1/(N-irs2) * Y(1:N-irs2,:).' * Y(irs2+1:N,:);
    end
end

%% CALCULATE SVD OF TOEPLITZ MATRIX AND ESTABLISH OBSERVABILITY AND CONTROLLABILITY MATRICES
[U,D,V]=svd(T);

disp(['*** Rank(D) = ' num2str(rank(D))])
if isempty(n)
    figure;
    stem(diag(D));
    disp('*** Please input the system orders in normal MATLAB format, e.g. 2:2:10 or [2,10,40].');
    n = input('System orders, n = ');
    if max(n)<l
        warning(['Maximum system order too small. Should be minimum equal to the number of channels. Adjusted to ' num2str(l) '.'])
        n=[n, l];
    end
end
if rank(D)<max(n)
    warning('System order reduced to rank of D matrix.')
    n=n(n<rank(D));
end

disp(['*** Maximum system order used is n_max = ' num2str(max(n))]);

%% COMPUTE STATE MATRIX A FROM DECOMPOSITION FOR EACH ORDER, AND ESTABLISH EIGENVALUES AND EIGENVECTORS
for j = 1:length(n)  %order indices, j
    o=n(j);          %the current order (jth value of i-array)
    D1=D(1:o,1:o);
    U1=U(:,1:o);
    V1=V(:,1:o);
    
    % Establish O and F from subset of singular values
    %     F = sqrt(D1)*V1.';    %Gamma, controllability
    %     G = F(:,end-l+1:end);
    O = U1*sqrt(D1);      %O, observability (T=identity matrix)
    C = O(1:l,:);         %Pick the l first values
    
    A = sqrt(D1)\U1.'*T2*V1/sqrt(D1); %state matrix A
    [Phi0,Lambda0] = eig(A);                    %eigenvectors and eigenvalues
    lambda0=log(diag(Lambda0))./dt;             %make vector and transform from discrete to continuous
    Phi0=C*Phi0;
    
    [~,ix]=unique(abs(lambda0));            %find index for unique absolute values
    lambda{j}=lambda0(ix);                  %keep the corresponding eigenvalues
    Phi{j}=Phi0(:,ix);                      %and the corresponding eigenvectors
end
order=n;