function stabPlot(warray,xiarray,orders,varargin)
%%STABPLOT Plot stabilization plot for w and xi.
% Use like this: stabPlot(warray,xiarray,orders,varargin)
%
% Knut Andreas Kvaale, 2014/15
%

p=inputParser;
addParameter(p,'criterion',[0.01,0.05]);
addParameter(p,'plots','both')
addParameter(p,'stablevel',1)
addParameter(p,'markers','.')      %first value is "ALL", second is "STABLE"
addParameter(p,'markersize',[36])  %first value is "ALL", second is "STABLE"
addParameter(p,'colors',{'r' 'k'})
parse(p,varargin{:})

criterion=p.Results.criterion;
plots=p.Results.plots;
stabLevel=p.Results.stablevel;
markers=p.Results.markers;
markersize=p.Results.markersize;
colors=p.Results.colors;

if length(markersize)==1
    markersize(2)=markersize(1);
end

if length(markers)==1
    markers(2)=markers(1);
end

wtol=criterion(1);
xitol=criterion(2);

%% INITIALIZE PLOT CONTAINERS
stableFreq=[];    %X
stableOrders=[];  %Y in stable plot
allFreq=[];
allOrders=[];

%% ESTABLISH FOR ORDER BELOW AND EQUAL TO STABLEVEL
for i = 1:stabLevel
    allFreq=[allFreq; warray{i}];%Hz?
    allOrders=[allOrders; ones(length(warray{i}),1)*orders(i)];
    allXi{i}=xiarray{i};
end

%% ESTABLISH FOR ORDER ABOVE STABLEVEL
for i = stabLevel+1:length(orders)
    w=warray{i};
    xi=xiarray{i};
    
    allFreq=[allFreq; w];
    allOrders=[allOrders; ones(length(w),1)*orders(i)];
    allXi{i}=xiarray{i};
    
    %STABLE POLES
    for m = 1:length(w)
        stab=0;
        for level=1:stabLevel
            wlast=warray{i-level};
            [dw,iwlast] = min(abs(w(m)-wlast));
            xilast=xiarray{i-level}(iwlast);
            dxi = xi(m)-xilast;
            if (dw/wlast(iwlast)<=wtol)&&(dxi/xilast<=xitol)
                stab=stab+1;
            else
                stab=0;
                break
            end
        end
        
        if  stab==stabLevel
            stableFreq=[stableFreq; w(m)];
            stableOrders=[stableOrders; orders(i)];
        end
    end
end

%% PLOT
leg={};
f=figure; hold on

if strcmp(plots,'both') || strcmp(plots,'all')
    a=scatter(allFreq,allOrders,markers(1));
    set(a,'sizedata',markersize(1),'MarkerEdgeColor',colors{1})
    leg{length(leg)+1} = 'All poles';
end

if strcmp(plots,'both') || strcmp(plots,'stable')
    b=scatter(stableFreq,stableOrders,markers(2));
    set(b,'sizedata',markersize(2),'MarkerEdgeColor',colors{2})
    leg{length(leg)+1} = 'Stable poles';
end

legend(leg)
title({[num2str(stabLevel) ' x stability '], ['Criteria: \Delta\omega/\omega < ' num2str(100*wtol) '% and \Delta\xi/\xi < ' num2str(100*xitol) '%']});
xlabel('Mode frequency')
ylabel('Order');
box

dcm=datacursormode(f);
datacursormode on

set(dcm,'updatefcn',{@infoCursor warray allXi orders})

function [output_txt] = infoCursor(obj,event_obj,warray,allXi,orders)
% Display the position of the data cursor
% obj          Currently not used (empty)
% event_obj    Handle to event object
% output_txt   Data cursor text string (string or cell array of strings).


pos = get(event_obj,'Position');

wsel = pos(1); %frequency
nsel = pos(2); %order
[~,isel] =min(abs(orders-nsel)); %find index of order
[~,msel]=min(abs(warray{isel}-wsel)); %find number of mode for this order
xisel= allXi{isel}(msel);

index = isel;
mode = msel;
save('stabPlotResults.mat','index','mode')

output_txt = {...    
    ['order = ',num2str(nsel,3)],...
    ' ',...
    ['w = ',num2str(wsel,4)],...
    ['xi = ', num2str(xisel*100,4) '%'],...
    ' ',...
    ['mode = ', num2str(msel,3)],...
    ['index = ', num2str(isel,3)]};
